/* ****************************************************************
 * Copyright Stanford University 1998,99 - All Rights Reserved
 ****************************************************************** 

 * Permission to use, copy, modify, and distribute this software 
 * and its documentation for any purpose is hereby granted without 
 * fee, provided that the above copyright notice appears in all copies
 * and that both the copyright notice, this permission notice, and 
 * the following disclaimer appear in supporting documentation, and 
 * that the name of Stanford University, not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * STANFORD UNIVERSITY, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS. IN NO EVENT SHALL STANFORD UNIVERSITY BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION 
 * OF CONTRACT, NEGLIGENCE OR OTHER ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


/*
 * The Sim Web Site : http://klamath.stanford.edu/tools/SIM
 * The SIM Mailing List: sim-simulator@lists.stanford.edu

 * Send mail to the above email address with "subscribe sim-simulator" in
 * the body of the message.
 *
 */

#include "sim.h"
#include "fabric.h"
#include "inputAction.h"
#include <string.h>


struct CrossbarFabricState {
  int maxCompare;   /* If set, compare size of match with max sized match. */
};

struct CrossbarStats {
  Stat matchSize;
  Stat matchCompare;
};

int resetMatrix(Switch *aSwitch );
extern int findSizeMaxMatch(Switch *aSwitch);
extern void printCell(FILE*, Cell*);

/* 
   Crossbar switch. 
   Using aSwitch->fabric.Xbar_matrix, take one cell from each scheduled
   input fifo and move to output fifo. 
   Only one cell may be taken from each input buffer.
   Only one cell may be added to each output buffer.
*/
int 
crossbar(action, aSwitch, argc, argv)
  FabricAction action;
Switch *aSwitch;
int argc;
char **argv;
{
  /* 
     For each output in the matrix for this switch we:
     remove one cell from the corresponding input buffer
     check to see if any more destinations for the cell
     if so, add it to the input buffer at its destination
     if not, destroy the cell
     */
  Cell *aCell;
  struct CrossbarFabricState *fabricState;
  struct CrossbarStats *fabricStats;

  int outputLineWidth = 1 ;	/* NOTE : specific to PBC switch architecture. */

  if(debug_fabric)
    printf("Fabric 'crossbar()' called by switch %d\n",aSwitch->switchNumber);

  switch(action)
    {
    case FABRIC_USAGE:
      fprintf(stderr, "Options for \"crossbar\" fabric:\n");
      fprintf(stderr, "  -m Compare match with maximum size\n");
      fprintf(stderr, "  -o numOutputLines per output. Default: 1\n");
      break;
    case FABRIC_INIT:
      {
	extern int opterr, optind;
	extern int optopt;
	extern char *optarg;
	int i,c;

	if( debug_fabric )
	  printf("     Fabric: INIT crossbar matrix\n");
	if( !aSwitch->fabric.Xbar_matrix )
	  {

	    /* Parse options */	
	    opterr=0;
	    optind=1;

	    fabricState = (struct CrossbarFabricState *)
	      malloc( sizeof(struct CrossbarFabricState) );
	    aSwitch->fabric.fabricState = fabricState;
	    fabricState->maxCompare = 0;

	    for(i=0; i<=argc; i++)
	      {
		c = getopt(argc, argv, "mo:");
		switch (c)
		  {
		  case 'm':
		    fabricState->maxCompare = 1;
		    break;
		  case 'o':
		    outputLineWidth = atoi(optarg);
		    break;
		  default:
		    optind++;
		    break;
		  }
	      }

		/* PBC : Overwrite output line width with the value of B. */
	    // if( aSwitch->numInternalBuffers != NONE )	outputLineWidth = aSwitch->numInternalBuffers ;
		  
		aSwitch->fabric.Xbar_numOutputLines = outputLineWidth ; 
	    /* Create switch routing matrix */
	    aSwitch->fabric.Xbar_matrix = (MatrixEntry *) malloc( sizeof(MatrixEntry) *
															  aSwitch->numOutputs *
															  aSwitch->fabric.Xbar_numOutputLines);

	    if( outputLineWidth != 1 )
	      printf("Crossbar output line width: %d\n", 
		     aSwitch->fabric.Xbar_numOutputLines);
	  }
	/* Init switch routing matrix */
	resetMatrix(aSwitch);

	break;
      }
    case FABRIC_EXEC:
      {
	struct List  *toFifo;
	struct Element	*anElement;
	int switchInput, switchOutput;
	int fabricOutput, iB, i, m, n ;
	int matchSize=0, maxSizeMatch=0, numFabricOutputs;
	int priority;

	if( debug_fabric )
	  printf("     Fabric: EXEC crossbar matrix\n");

	fabricState = aSwitch->fabric.fabricState;
	fabricStats = aSwitch->fabric.fabricStats;

	if( fabricState->maxCompare )
	  maxSizeMatch = findSizeMaxMatch(aSwitch);

	numFabricOutputs = aSwitch->numOutputs * aSwitch->fabric.Xbar_numOutputLines;
	
	for(fabricOutput=0; fabricOutput<numFabricOutputs; fabricOutput++)
	{
		switchOutput = fabricOutput/aSwitch->fabric.Xbar_numOutputLines;
		
		/* B = 1 */
		if( aSwitch->numInternalBuffers == NONE )
		{
			/* Determine "switchInput" that this fabricOutput will be connected to. */
			switchInput = aSwitch->fabric.Xbar_matrix[fabricOutput].input;
			/* Determine "aCell" that is to be transferred. */
			aCell = aSwitch->fabric.Xbar_matrix[fabricOutput].cell;
			
			/* Transmit the cell with the classic way ! */
			if( switchInput != NONE )
			{
				aCell = (aSwitch->inputAction)(INPUTACTION_TRANSMIT, aSwitch, switchInput, (Cell *)aCell, fabricOutput);
				
				/* ******************************* */
				/* Determine output buffer s FIFO. */
				/* ******************************* */
				priority = aCell->priority;
				toFifo = aSwitch->outputBuffer[switchOutput]->fifo[priority];   /* TL */
				
				/* Store the cell to output buffer FIFO. */
				anElement = createElement(aCell);
				addElement(toFifo, anElement);			  
				matchSize++;
				
				if( debug_fabric )	printf("     Fabric: Update stats\n");
				/* Update latency stats for this cell */
				latencyStats(LATENCY_STATS_ARRIVE_FABRIC, NULL, aCell);
				latencyStats(LATENCY_STATS_ARRIVE_OUTPUT, NULL, aCell);
				if(debug_fabric)	printf("Cell from input %d to output %d priority %d at time %lu\n", switchInput, fabricOutput, priority, now);
				if(debug_fabric)	printCell(stdout, aCell);
			}
		}
		else	/* B != 1 ---> PBC architecture */
		{
			/* Determine "switchInput" that this fabricOutput will be connected to. */
			/* Determine "aCell" that is to be transferred. */
			for(i=0; i<aSwitch->numInputs; i++)
				if( aSwitch->inputBuffer[i]->grantQueue[fabricOutput] == 1 )
				{
					switchInput = i ;
					aCell = (Cell *)aSwitch->inputBuffer[i]->fifo[fabricOutput]->head->Object ;
					
					/* Transmit the cell according to PBC flow control ! */
					if( switchInput != NONE )
					{
						aCell = (aSwitch->inputAction)(INPUTACTION_TRANSMIT, aSwitch, switchInput, (Cell *)aCell, fabricOutput);
						
						/* ******************************* */
						/* Determine output buffer s FIFO. */
						/* ******************************* */
						for( iB=0; iB<aSwitch->numInternalBuffers; iB++ )
							if( aSwitch->outputBuffer[switchOutput]->fifo[iB]->head == NULL )	
							{ toFifo = aSwitch->outputBuffer[switchOutput]->fifo[iB] ; break ; }
						if( toFifo == NULL )	fprintf(stderr, "ERROR : The flow control of the scheduling algorithm that checks the buffers availability is WRONG ! \n") ;
						/* ******************************* */
						
						/* Store the cell to output buffer FIFO. */
						anElement = createElement(aCell);
						addElement(toFifo, anElement);			  
						matchSize++;
						
						/* ******** */
						/* AFdropPR */
						/* ******** */
						/* ON-service VOQ : (1) if cf(i,j)>1 and (2) the 1st cell of the frame has been transmitted :S . */
						if( aSwitch->inputBuffer[switchInput]->cf[switchOutput] > 1 )	
						{
							aSwitch->inputBuffer[switchInput]->service[switchOutput] = 1 ;
							if( debug_fabric )	fprintf(stdout, "--- ALARM(%ld) : Circuit (%d,%d) is now opened ! --- \n", now, switchInput, switchOutput) ;
							aSwitch->outputBuffer[switchOutput]->frame = switchInput ;
							aSwitch->inputBuffer[switchInput]->capture = switchOutput ;
						}
						/* OFF-service VOQ : if the last cell of the frame has been sent ! OR no cell of the frame has been sent to Bs (flow control does this by default). */
						if( (aSwitch->inputBuffer[switchInput]->cf[switchOutput] == 1) && (aSwitch->inputBuffer[switchInput]->service[switchOutput] == 1) )	
						{
							aSwitch->inputBuffer[switchInput]->service[switchOutput] = 0 ;
							if( debug_fabric )	fprintf(stdout, "--- ALARM(%ld) : Circuit (%d,%d) is now closed ! --- \n", now, switchInput, switchOutput) ;
							aSwitch->outputBuffer[switchOutput]->frame = NONE ;
							aSwitch->inputBuffer[switchInput]->capture = NONE ;
						}
						/* The VOQ is in a ON-service state till the last cell its frame is serviced. */
						/* VS. Note that for frame sizes of 1 cell, the associated VOQ is OFF-service during the matching of its 1 cell frame. */
						// if( aSwitch->inputBuffer[switchInput]->cf[fabricOutput] == 0 )	aSwitch->inputBuffer[switchInput]->service[fabricOutput] = 0 ;
						
						/* cf(i,j) -- each time a cell is serviced. */
						aSwitch->inputBuffer[switchInput]->cf[switchOutput] -- ;
						
						/* ******** */
						/* EXdropPR */
						/* ******** */
						if( aSwitch->inputBuffer[switchInput]->fifo[switchOutput]->number > 1 )	
						{
							aSwitch->inputBuffer[switchInput]->exhaustOutput = switchOutput ;
							if( debug_fabric )	fprintf(stdout, "--- ALARM(%ld) : Circuit (%d,%d) is now opened ! --- \n", now, switchInput, switchOutput) ;
						}
						else	
						{
							aSwitch->inputBuffer[switchInput]->exhaustOutput = NONE ;
							aSwitch->outputBuffer[switchOutput]->exhaustInput = NONE ;
							if( debug_fabric )	fprintf(stdout, "--- ALARM(%ld) : Circuit (%d,%d) is now closed ! --- \n", now, switchInput, switchOutput) ;
						}
						if( aSwitch->inputBuffer[switchInput]->exhaustOutput != NONE )	aSwitch->outputBuffer[switchOutput]->exhaustInput = switchInput ;
						
						
						if( debug_fabric )	printf("     Fabric: Update stats\n");
						/* Update latency stats for this cell */
						latencyStats(LATENCY_STATS_ARRIVE_FABRIC, NULL, aCell);
						latencyStats(LATENCY_STATS_ARRIVE_OUTPUT, NULL, aCell);
						if(debug_fabric)	printf("Cell from input %d to output %d priority %d at time %lu\n", switchInput, fabricOutput, priority, now);
						if(debug_fabric)	printCell(stdout, aCell);
					}
				}
		}
		  
	}
		  
	  if( debug_fabric )
	  {
		  fprintf(stdout, "--- A D A P T (cont.) --- \n") ;
		  for( m=0; m<aSwitch->numInputs; m++ )
		  {
			  for( n=0; n<aSwitch->numOutputs; n++)
				  fprintf(stdout, "[%d,%d] \t", aSwitch->inputBuffer[m]->cf[n], aSwitch->inputBuffer[m]->service[n]) ;
			  fprintf(stdout, "\n") ;
		  }
	  }
		  
	updateStat(&fabricStats->matchSize, (long) matchSize, now);

	if( fabricState->maxCompare )
	  {
	    updateStat(&fabricStats->matchCompare, 
		       (long) maxSizeMatch-matchSize, now);
	  }
	resetMatrix(aSwitch);

	break;
      }
    case FABRIC_STATS_PRINT:
      {
	fabricStats = aSwitch->fabric.fabricStats;
	fabricState = aSwitch->fabric.fabricState;

	printf("\n");
	printf("  Crossbar statistics\n");
	printf("  -------------------\n");
	printf("                       Avg       SD     Number\n");
	printf("                     -------------------------\n");
	printStat(stdout, "Match Size:         ", 
		  &fabricStats->matchSize);
	if( fabricState->maxCompare )
	  printStat(stdout, "Match diff from Max:", 
		    &fabricStats->matchCompare);
	break;
      }
    case FABRIC_STATS_INIT:
      {
	fabricState = aSwitch->fabric.fabricState;
	fabricStats = aSwitch->fabric.fabricStats;
	if( !aSwitch->fabric.fabricStats )
	  {
	    /* Create and init fabric stats */
	    fabricStats= (struct CrossbarStats *)
	      malloc(sizeof(struct CrossbarStats));
	    aSwitch->fabric.fabricStats = (void *) fabricStats;
	  }
	initStat(&fabricStats->matchSize, STAT_TYPE_AVERAGE, now);
	initStat(&fabricStats->matchCompare, STAT_TYPE_AVERAGE, now);
	enableStat(&fabricStats->matchSize);
	if( fabricState->maxCompare )
	  enableStat(&fabricStats->matchCompare);
	break;
      }
    default:
      fprintf(stderr, "\nCommand not implemented in crossbar.c\n");
      exit(2);
    }
  if( debug_fabric )
    printf("     Fabric: Completed\n");

  return (0);
}

int resetMatrix( aSwitch )
  Switch *aSwitch;
{
  int numFabricOutputs;
  
  numFabricOutputs = aSwitch->numOutputs*aSwitch->fabric.Xbar_numOutputLines;
  
  if( debug_fabric )
    printf("     Fabric: Reset crossbar matrix\n");
  memset(aSwitch->fabric.Xbar_matrix, NONE, numFabricOutputs*sizeof(MatrixEntry));
  return (0);
}
