/* ****************************************************************
 * Copyright Stanford University 1998,99 - All Rights Reserved
 ****************************************************************** 

 * Permission to use, copy, modify, and distribute this software 
 * and its documentation for any purpose is hereby granted without 
 * fee, provided that the above copyright notice appears in all copies
 * and that both the copyright notice, this permission notice, and 
 * the following disclaimer appear in supporting documentation, and 
 * that the name of Stanford University, not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * STANFORD UNIVERSITY, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS. IN NO EVENT SHALL STANFORD UNIVERSITY BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION 
 * OF CONTRACT, NEGLIGENCE OR OTHER ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * The Sim Web Site : http://klamath.stanford.edu/tools/SIM
 * The SIM Mailing List: sim-simulator@lists.stanford.edu

 * Send mail to the above email address with "subscribe sim-simulator" in
 * the body of the message.
 *
 */

#include <sys/time.h>
#include <math.h>
#include "sim.h"

#include "algorithm.h"
#include "fabric.h"
#include "traffic.h"
#include "inputAction.h"
#include "outputAction.h"


#ifdef _SIM_
#include "graph.h" // SG
#endif // _SIM_


#define DEFAULT_SIMULATION_LENGTH 100
#define STOP_THRESHOLD 0.0001 
#define NUM_TIMES_MUST_MEET_STOP_CONDITION 10
#define CHECK_STOP (simulationLength/50+1)

#define CHECK_MEMORY_USAGE (simulationLength/50+1)
#define MAX_ALLOWED_MEMORY_USAGE 20000000

/*** static char ident[] = "@(#)ATM Switch Simulator Revision 2.31"; ***/


/**********************static functions **************/
static time_t timeDiff();
static void headerInformation();
static int memoryUsage();
static int checkStopCondition();

/******functions defined here and also used elsewhere***********/

void FatalError();
void automateLoadDelayCurve( Switch * ) ;

/******functions defined elsewhere and used here***********/
extern void parseConfigurationFile();
extern void switchStats();/* from switchStats.c */
extern void burstStats(); /* from switchStats.c */
extern void printInputFifoLevels(); /* from debug.c */
extern void printMatrix();/* from debug.c */
extern int getAllocAmount();/* from circBuffer.c */
extern void gethostname(); /* some platforms e.g. SunOS complain if this is not declared */
extern void instantaneousOutputData( Switch * ) ;

/*****************************************************/
/********** Universal time measured in cells. ********/
/*****************************************************/
long now=0;
long resetStatsTime=NONE;
unsigned long globalSeed=0;

#ifdef _SIM_
struct GRAPH graphTool; // SG
#endif // _SIM_


Switch **switches;
int numSwitches=1;
long simulationLength=DEFAULT_SIMULATION_LENGTH; 

/*#ifdef use_gnuplot
	double inputLoad = 0.0 ;	// equals to the sum of all input ports loads, used by sim.c:automateLoadDelayCurve.
#endif*/
// char	**inputTrafficModel ;
double	*weights ;


/**************************************************/
/*************main function ***********************/
/**************************************************/
int main(argc, argv)
  int argc;
char **argv;
{
  Switch *aSwitch;
  int switchNumber;
  int input,output;

  char c;
  int i;
  extern char *optarg;
  extern int optind;
  extern int opterr;

  int simStopped=CONTINUE_SIMULATION;
  int fflag=0;
  int pflag=0;
  int checkPeriod=0;
  char *configFilename=(char *) NULL;

  /* XXX Temp for varying internal speeds */
  int trafficPeriod=1;
  int fabricPeriod=1;
  int outputPeriod=1;
  

#ifdef _SIM_

  /************************************************************/
    /* SIMGRAPH related global variables (SG)                   */
    /************************************************************/

    int simGraphSocketPeriod=10; // Amr
    int simGraphEnabled = 0; // Amr
    int graphTCPPort = DEFAULT_TCP_PORT;
    int graphNumRequests = DEFAULT_MAX_NUM_REQS;
    extern long simGraphPauseTime;

#endif //_SIM_

  /************************************************************/
  /* For long simulation runs: check that simulation does not */
  /* use too much memory.                                     */
  /************************************************************/
  int memUsed=0, memMaxUsed=0;

  /************************************************************/
  /******* Used to measure how long the simulation runs *******/
  /************************************************************/
  time_t startTime, stopTime;

  /*********************************************************/
  /*********** Print header information to stdout **********/
  /*********************************************************/
  headerInformation(argc,argv);

  /*********************************************************/
  /*********** Note time that simulation starts ************/
  /*********************************************************/
  time(&startTime);
  /* srand48(startTime); */

  /*********************************************************/
  /************** Parse top level user options *************/
  /*********************************************************/
  opterr=0;
  optind=1;
  for(i=0; i<=argc; i++)
    {
#ifdef _SIM_

      c = getopt(argc, argv, "vhf:r:u:l:pt:s:e:GP:R:S:"); 
		// SG: Added options: 
        //     G:Enable SIMGRAPH, P:SIMGRAPH TCP port,
        //     S:Socket Polling period, R: Number of requests.
#else
      c = getopt(argc, argv, "vhf:r:u:l:pt:s:e:"); 
#endif // _SIM_
      switch (c)
	{

#ifdef _SIM_

   /**************************************************/
     /* G, P, R, S are SIMGRAPH specific options (SG)  */
     /**************************************************/
 	case 'G': /* SimGraph Enable Flag, Default is disabled */
     simGraphEnabled = 1;
     printf("SimGraph is enabled\n");
     break;
 	case 'P': /* Graphsim TCP port: DEFAULT_TCP_PORT */
     graphTCPPort = atoi(optarg);
     printf("SimGraph TCP port number = %d\n", graphTCPPort);
     break;
 	case 'R': /* Graphsim Num Requests: DEFAULT_MAX_NUM_REQS */
     graphNumRequests = atoi(optarg);
     printf("SimGraph number of requests = %d\n", graphNumRequests);
     break;
 	case 'S': /* SimGraph Socket Period, default is 10 */
     simGraphSocketPeriod = atoi(optarg);
     printf("SimGraph socket period = %d\n", simGraphSocketPeriod);
     break;

#endif // _SIM_

	case 'l':	/* Length of simulation in cells: default=100 */
	  simulationLength =  atoi(optarg);
	  break;
	case 'r':	/* Reset Stats time */
	  resetStatsTime = atol(optarg);
	  break;
	case 't':	/* Traffic Period */
	  trafficPeriod = atoi(optarg);
	  printf("Traffic period: %d\n", trafficPeriod);
	  break;
	case 's':	/* Switch/Scheduling/Fabric Period */
	  fabricPeriod = atoi(optarg);
	  printf("Fabric period: %d\n", fabricPeriod);
	  break;
	case 'e':	/* Output/Exit Period */
	  outputPeriod = atoi(optarg);
	  printf("Output period: %d\n", outputPeriod);
	  break;
	case 'f':	/* Configuration filename */
	  fflag = 1;
      configFilename = (char *) malloc(sizeof(char)*(strlen(optarg)+1));
	  strcpy(configFilename, optarg);
	  break;
	case 'p':	/* Inidicate progress as simulation proceeds */
	  pflag = 1;
	  break;
	case 'u':       /* Sets random seed for rand generators */
	  globalSeed = atoi(optarg);
      if(globalSeed == 0)
      {
	    time((time_t *) &globalSeed);
	    printf("Randomly selected seed: ");
      }
	  printf("Simulation using seed %lu\n",globalSeed);
	  break;
	case 'v': /* Print version number */
      printf("Sim v2.35, Released Sun, April 8 2001 \n");
      exit(0);
	case 'h':
	  fprintf(stderr, "usage: %s \n", argv[0]);
	  fprintf(stderr, "GENERAL options:\n");
	  fprintf(stderr, "    -h This help message \n");
	  fprintf(stderr, "    -l simulationlength. Default: 100\n");
	  fprintf(stderr, "    -r resetStatsTime. Default: simLength/10\n");
	  fprintf(stderr, "    -t Traffic Period Default: 1 \n");
	  fprintf(stderr, "    -s Fabric Period Default: 1 \n");
	  fprintf(stderr, "    -e Exit or Output Period Default: 1 \n");
	  fprintf(stderr, "    -p Indicate progress\n");
	  fprintf(stderr, "    -f configFilename.Default: none\n");
	  fprintf(stderr, "    -u globalSeed. Default: None \n");  
	  fprintf(stderr, "\nTraffic Models:\n");
	  fprintf(stderr, "----------------------------------------\n");
	  FOREACH_FUNCTION(trafficTable, TRAFFIC_USAGE);
	  fprintf(stderr, "\nInput Actions:\n");
	  fprintf(stderr, "----------------------------------------\n");
	  FOREACH_FUNCTION(inputActionTable, INPUTACTION_USAGE);
	  fprintf(stderr, "\nSwitch scheduling algorithms:\n");
	  fprintf(stderr, "----------------------------------------\n");
	  FOREACH_FUNCTION(algorithmTable, SCHEDULING_USAGE);
	  fprintf(stderr, "\nOutput Actions:\n");
	  fprintf(stderr, "----------------------------------------\n");
	  FOREACH_FUNCTION(outputActionTable, OUTPUTACTION_USAGE);
	  fprintf(stderr, "\nSwitch Fabrics:\n");
	  fprintf(stderr, "----------------------------------------\n");
	  FOREACH_FUNCTION(fabricTable, FABRIC_USAGE);
#ifdef _SIM_

        /**************************************************/
        /* G, P, R, S are SIMGRAPH specific options (SG)  */
        /**************************************************/
        fprintf(stderr,"\nSimGraph options:\n");
        fprintf(stderr,"    -G: Enable SimGraph. Default: disabled\n");
        fprintf(stderr,"    -P: SimGraph TCP port number.\n");
        fprintf(stderr,"    -R: SimGraph Number of Requests.\n");
        fprintf(stderr,"    -S: SimGraph Socket Polling Period. Default: 10\n");

#endif // _SIM_
	    
	  exit(0);
	default:
	  break;
	}
				
    }


  /********************************************************/
  /***** Parse config file and create switches ************/
  /********************************************************/
  if(!fflag)
    FatalError("Must specify -f option configuration file.");

  parseConfigurationFile( configFilename );

  if( resetStatsTime == NONE )
    resetStatsTime = simulationLength / 2;
  printf("ResetStatsTime %ld\n", resetStatsTime);

#ifdef _SIM_

  /************************************************************/
    /***** Initialize SIMGRAPH  (SG)                ************/
    /************************************************************/
    if( simGraphEnabled )
        {
            char aux[10];
            graph_init( &graphTool );
            sprintf(aux, "%d", graphNumRequests);
            graph_config( &graphTool, 0, 0, "maxNumReqs", aux);
            sprintf(aux, "%d", graphTCPPort);
            graph_config( &graphTool, 0, 0, "TCPportNum", aux);
            graph_start( &graphTool );
        }

#endif // _SIM_


  /**************************************************************/
  /******************* START SIMULATION *************************/
  /**************************************************************/
  for( now=0; (now<simulationLength) && (simStopped==CONTINUE_SIMULATION); 
       now++ )
    {
      if(debug_sim)
	{
	  printf("\nTIME %ld\n", now);
	  printf("---------\n");
	}


      /* Is it time to reset statistics? */
      if( now == resetStatsTime || now == 0 )
	{
	  for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
	    {
	      aSwitch = switches[switchNumber];
	      printf("Resetting stats for switch: %d at time %lu\n", 
		     switchNumber, now);
	      switchStats(SWITCH_STATS_RESET, aSwitch); 
	      latencyStats(LATENCY_STATS_SWITCH_INIT, aSwitch, NULL);
	      for(output=0; output<aSwitch->numOutputs; output++)
		burstStats(BURST_STATS_RESET, aSwitch, output); 
	      (aSwitch->scheduler.schedulingAlgorithm)(SCHEDULING_INIT_STATS, aSwitch, aSwitch, NULL);
	      (aSwitch->fabric.fabricAction)(FABRIC_STATS_INIT, aSwitch);
	    }
	}


      /******************************************************/
      /********* Determine new traffic **********************/
      /************** for every switch **********************/
      /******************************************************/
      for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
	{
	  if(now%trafficPeriod) continue;

	  if(debug_sim)
	    {
	      printf("SWITCH %d:\n", switchNumber);
	      printf("---------\n");
	    }
	  aSwitch = switches[switchNumber];

	  /***********************************************************/
	  /* The following is temporary until separate traffic units */
	  /* are implemented: */
	  /***********************************************************/
	  /* Determine new traffic for each input of switch */
	  if(debug_sim) printf("	New traffic\n");

	  for( input=0; input<aSwitch->numInputs; input++ )
	    {
	      if( (aSwitch->inputBuffer[input]->trafficModel)
		  (TRAFFIC_GENERATE, aSwitch, input, NULL, NULL) == 
		  STOP_SIMULATION)
		simStopped = STOP_SIMULATION;
	    }
	  if(debug_sim) printf("	Finished New traffic\n");
	}


		/* Slackness. */
		/*#ifdef	use_gbvoq
		 struct Element	*gbvoqElement ;
		 Cell			*gbvoqCell ;
		 int				gbvoq_in, gbvoq_out, inputThread, outputCushion, slackness ;
		 FILE			*gbvoq ;
		 char			*gbvoq_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/gbvoq.dat" ;
		 #endif
		 
		 if( use_gbvoq )
		 for(gbvoq_in=0; gbvoq_in<aSwitch->numInputs; gbvoq_in++)
		 if(aSwitch->inputBuffer[gbvoq_in]->inputPriorityList->head != NULL)
		 EACH_ELEMENT(aSwitch->inputBuffer[gbvoq_in]->inputPriorityList, gbvoqElement)
		 {
		 gbvoqCell = (Cell *)gbvoqElement->Object ;
		 inputThread = priorObjectsInList(aSwitch->inputBuffer[gbvoq_in]->inputPriorityList, gbvoqCell) ;
		 gbvoq_out = gbvoqCell->vci ;
		 outputCushion = priorObjectsInList(aSwitch->outputBuffer[gbvoq_out]->outputPriorityList, gbvoqCell) ;
		 slackness = outputCushion-inputThread ;
		 fprintf(stderr, "time %ld ... in:[%d] out:[%d] ... OC=%d(%ld), IT=%d(%ld), L(%lu)=%d \n", now, gbvoq_in, gbvoq_out, outputCushion, gbvoqCell->commonStats.oqDepartTime, inputThread, gbvoqCell->commonStats.arrivalTime, gbvoqCell->commonStats.ID, slackness) ;
		 }*/
		

      /**********************************************************/
      /******** Execute fabric action for all switches **********/
      /** Fabric transfers cells from input to output queues ****/
      /**********************************************************/
      for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
	{

	  if(now%fabricPeriod) continue;
	
	  aSwitch = switches[switchNumber];	
		
		/* $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */
		/*										LOG_FILE														*/
		/* $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */
#ifdef	use_slackness
	Cell			*dtcCell ;
	int				dtc_in, dtc_out, dtc_iB, dtc_i, dtc_b, inputThread, outputCushion, dtcSlackness ;
	char			*slackness_hol_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_ports/slackness_hol.dat" ;
	FILE			*slackness_hol ;
#endif
		
		if( use_slackness/*aSwitch->numInternalBuffers != NONE*/ )
		{
			if( now >= resetStatsTime )
			{
				if( (slackness_hol = fopen(slackness_hol_file, "a")) == NULL )
					{ fprintf(stderr, "ERROR : Cannot open file 'slackness_hol.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'slackness HoL data' are based on the filename - '%s' \n", slackness_hol_file) ; exit( EXIT_FAILURE ) ; }
				
				for( dtc_out=0; dtc_out<aSwitch->numOutputs; dtc_out++ )
				{
					fprintf(slackness_hol, "%ld\t%d\t", now, dtc_out) ;
					dtc_i = 0 ;
					for( dtc_in=0; dtc_in<aSwitch->numInputs; dtc_in++ )
					{
						if( aSwitch->inputBuffer[dtc_in]->fifo[dtc_out]->head != NULL )
						{
							dtcCell = (Cell *)aSwitch->inputBuffer[dtc_in]->fifo[dtc_out]->head->Object ;
							outputCushion = priorObjectsInList(aSwitch->outputBuffer[dtc_out]->outputPriorityList, dtcCell) ;
							inputThread = priorObjectsInList(aSwitch->inputBuffer[dtc_in]->inputPriorityList, dtcCell) ;
							dtcSlackness = outputCushion - inputThread ;
							if( dtcSlackness < 0 )	dtc_i ++ ;
						}
					}
					fprintf(slackness_hol, "%d\t", dtc_i);
					dtc_b = 0 ;
					for( dtc_iB=0; dtc_iB<aSwitch->numInternalBuffers; dtc_iB++ )
					{
						if( aSwitch->outputBuffer[dtc_out]->fifo[dtc_iB]->head != NULL )	dtc_b ++ ;
					}
					fprintf(slackness_hol, "%d", dtc_b);
					fprintf(slackness_hol, "\n") ;
				}
				
				if( fclose(slackness_hol) )	{ fprintf(stderr, "ERROR : Cannot close file 'slackness_hol.dat' ... \n") ; }
			}
		}
		
		
#ifdef	measure_instantaneous_hol
		int		i_in, i_out ;
		// int row, column, i, j, k ;
		// int		**instantHoL_perInput ;
		// int		*instantHoL_perOutput ;
		// FILE	*instantaneous_hol ;
		// char	*instantaneous_hol_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/instantaneous_hol.dat" ;
		Cell	*instantHoLCell ;
		char	*instantaneous_hol_weights_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/instantaneous_hol_weights.dat" ;
		FILE	*instantaneous_hol_weights ;
#endif
		/* Measure instantaneous number of HoL cells and save data to '_debug/_dats/instantaneous_hol.dat'. */
		if( measure_instantaneous_hol && (now>=resetStatsTime) )
		{
			/*instantHoL_perInput = (int **)malloc(aSwitch->numInputs*sizeof(int)) ;
			 if( instantHoL_perInput == NULL )	FatalError("measure_instantaneous_hol@sim.c: Malloc failed.\n") ;
			 instantHoL_perOutput = (int *)malloc(aSwitch->numOutputs*aSwitch->numPriorities*sizeof(int)) ;	
			 if( instantHoL_perOutput == NULL )	FatalError("measure_instantaneous_hol@sim.c: Malloc failed.\n") ;
			 for( i=0; i<aSwitch->numInputs; i++ )
			 {
			 instantHoL_perInput[i] = (int *)malloc(aSwitch->numOutputs*aSwitch->numPriorities*sizeof(int)) ;
			 if( instantHoL_perInput[i] == NULL )	FatalError("measure_instantaneous_hol@sim.c: Malloc failed.\n") ;
			 }
			 for( i=0; i<aSwitch->numInputs; i++ )
			 for( j=0; j<aSwitch->numOutputs*aSwitch->numPriorities; j++ )
			 instantHoL_perInput[i][j] = 0 ;
			 for( k=0; k<aSwitch->numOutputs*aSwitch->numPriorities; k++ )	instantHoL_perOutput[k] = 0 ;
			 
			 if( (instantaneous_hol = fopen(instantaneous_hol_file, "a")) == NULL )
			 { fprintf(stderr, "ERROR : Cannot open file 'instantaneous_hol.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'instantaneous HoL data' are based on the filename - '%s' \n", instantaneous_hol_file) ; exit( EXIT_FAILURE ) ; }
			 
			 fprintf(instantaneous_hol, "%ld \t", now) ;
			 for( i_in=0; i_in<aSwitch->numInputs; i_in++ )
			 for( i_out=0; i_out<aSwitch->numOutputs*aSwitch->numPriorities; i_out++ )
			 if( aSwitch->inputBuffer[i_in]->fifo[i_out]->number > 0 )
			 instantHoL_perInput[i_in][i_out] += 1 ;
			 for( column=0; column<aSwitch->numOutputs*aSwitch->numPriorities; column++ )
			 {
			 for( row=0; row<aSwitch->numInputs; row++ )
			 instantHoL_perOutput[column] += instantHoL_perInput[row][column] ;
			 fprintf(instantaneous_hol, "%d ", instantHoL_perOutput[column]) ;
			 }
			 fprintf(instantaneous_hol, "\n") ;
			 // NOTE : There is NO support for multicast traffic yet.
			 
			 if( fclose(instantaneous_hol) )	{ fprintf(stderr, "ERROR : Cannot close file 'instantaneous_hol.dat' ... \n") ; }*/
			
			/* Measure instantaneous weights of HoL cells and save data to '_debug/_dats/instantaneous_hol_weights.dat'. */
			if( (instantaneous_hol_weights = fopen(instantaneous_hol_weights_file, "a")) == NULL )
				{ fprintf(stderr, "ERROR : Cannot open file 'instantaneous_hol_weights.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'instantaneous HoL weights data' are based on the filename - '%s' \n", instantaneous_hol_weights_file) ; exit( EXIT_FAILURE ) ; }
			
			for( i_out=0; i_out<aSwitch->numOutputs*aSwitch->numPriorities; i_out++ )
			{
				fprintf(instantaneous_hol_weights, "%ld \t %d \t ", now, i_out) ;
				for( i_in=0; i_in<aSwitch->numInputs; i_in++ )
					if( aSwitch->inputBuffer[i_in]->fifo[i_out]->head != NULL )
					{
						instantHoLCell = (Cell *)aSwitch->inputBuffer[i_in]->fifo[i_out]->head->Object ;
						fprintf(instantaneous_hol_weights, "%.3g ", (double)(now/(float)instantHoLCell->commonStats.oqDepartTime)) ;
					}
					else	fprintf(instantaneous_hol_weights, "0.0 ") ;
				fprintf(instantaneous_hol_weights, "\n") ;
			}
			
			if( fclose(instantaneous_hol_weights) )	{ fprintf(stderr, "ERROR : Cannot close file 'instantaneous_hol_weights.dat' ... \n") ; }
			
			/*free(instantHoL_perOutput) ;
			 for( i=0; i<aSwitch->numInputs; i++ )	free(instantHoL_perInput[i]) ;
			 free(instantHoL_perInput) ;*/
		}
		
#ifdef	measure_instantaneous_output
		int				out, prio ; 
		// int no ;
		struct Element	*iElement ;
		// struct List		*outFifo ;
		// instantOut = (int *)malloc(simulationLength*aSwitch->numOutputs*sizeof(int)) ;
		// if( instantOut == NULL )	FatalError("defaultOutputAction(): Malloc failed.\n") ;
		// FILE			*instantaneous_output ;
		// char			*instantaneous_output_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/instantaneous_output.dat" ;
		Cell			*instantOutCell ;
		char			*instantaneous_out_weights_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/instantaneous_out_weights.dat" ;
		FILE			*instantaneous_out_weights ;
#endif
		
		/* Measure instantaneous output port occupancy and save data to '_debug/_dats/instantaneous_output.dat'. */
		if( measure_instantaneous_output && (now>=resetStatsTime) )
		{
			/*if((instantaneous_output = fopen(instantaneous_output_file, "a")) == NULL)
			 { fprintf(stderr, "ERROR : Cannot open file 'instantaneous_output.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'instantaneous output data' are based on the filename - '%s' \n", instantaneous_output_file) ; exit( EXIT_FAILURE ) ; }
			 
			 fprintf(instantaneous_output, "%ld \t", now) ;
			 for( out=0; out<aSwitch->numOutputs; out++ )
			 {
			 // outFifo = aSwitch->outputBuffer[out]->fifo[DEFAULT_PRIORITY]->number ;
			 // instantOut[now*aSwitch->numOutputs + out] = outFifo->number ;
			 fprintf(instantaneous_output, "%d ", aSwitch->outputBuffer[out]->fifo[DEFAULT_PRIORITY]->number) ;
			 }
			 fprintf(instantaneous_output, "\n") ;
			 
			 if( fclose(instantaneous_output) )	{ fprintf(stderr, "ERROR : Cannot close file 'instantaneous_output.dat' ... \n") ; }*/
			
			/* Measure instantaneous weights of output cells and save data to '_debug/_dats/instantaneous_output_weights.dat'. */
			if((instantaneous_out_weights = fopen(instantaneous_out_weights_file, "a")) == NULL)
			{ fprintf(stderr, "ERROR : Cannot open file 'instantaneous_out_weights.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'instantaneous out weights data' are based on the filename - '%s' \n", instantaneous_out_weights_file) ; exit( EXIT_FAILURE ) ; }
			
			for( out=0; out<aSwitch->numOutputs; out++ )
			{
				fprintf(instantaneous_out_weights, "%ld \t %d \t ", now, out) ;
				for( prio=0; prio<aSwitch->numPriorities; prio++ )
					if( aSwitch->outputBuffer[out]->fifo[prio]->head != NULL )
					{
						EACH_ELEMENT(aSwitch->outputBuffer[out]->fifo[prio], iElement)
						{
							instantOutCell = (Cell *)iElement->Object ;
							fprintf(instantaneous_out_weights, "%.3g ", (double)(now/(float)instantOutCell->commonStats.oqDepartTime)) ;
						}
					}
					else	if( prio == (aSwitch->numPriorities-1) )	{ fprintf(instantaneous_out_weights, "0.0 ", now, out) ; }
				fprintf(instantaneous_out_weights, "\n") ;
			}
			
			if( fclose(instantaneous_out_weights) )	{ fprintf(stderr, "ERROR : Cannot close file 'instantaneous_out_weights.dat' ... \n") ; }
		}
		/* $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */

	  for( input=0; input<aSwitch->numInputs; input++ )
		(aSwitch->inputAction)(INPUTACTION_PERCELL_CHECK, aSwitch, input, NULL);

	  /***************************************************/
	  /********  Execute Scheduling Algorithm ************/
	  /***************************************************/

	  if(debug_sim)	
	    {
	      printf("Levels BEFORE scheduling\n");
	      printInputFifoLevels( aSwitch );
	    }
		
	  if(debug_sim) printf("	Execute Scheduling Algorithm\n");
	  /* Configure switch: returns configuration matrix */
	  (aSwitch->scheduler.schedulingAlgorithm)(SCHEDULING_EXEC, aSwitch);

	  if(debug_sim)	
	    {
	      printf("Matrix AFTER scheduling\n");
	      printMatrix( aSwitch );
	    }

	  if(debug_sim)
	    (aSwitch->scheduler.schedulingAlgorithm)
	      (SCHEDULING_REPORT_STATE,aSwitch);

	  if( checkPeriod )
	    (aSwitch->scheduler.schedulingAlgorithm)
	      (SCHEDULING_CHECK_STATE_PERIOD, aSwitch); 

	  if(debug_sim) printf("	Switch fabric execution\n");
	  /* Transfer cells into/out of switch fabric. */
	  (aSwitch->fabric.fabricAction)(FABRIC_EXEC, aSwitch, NULL, NULL);
		
	}


      /**********************************************************/
      /******** Execute output actions for all switches *********/
      /**********************************************************/
      for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
	{
	  if(now%outputPeriod) continue;

	  aSwitch = switches[switchNumber];	
	
	  if(debug_sim) printf("	Switch output execution\n");
	  /* Transfer cells to next switch or destroy. */
	  (aSwitch->outputAction)(OUTPUTACTION_EXEC, aSwitch);
	}
		
      /**********************************************************/
      /******* Per celltime statistics for each switch. *********/
      /**********************************************************/
      for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
	{
	  aSwitch = switches[switchNumber];	
	  switchStats( SWITCH_STATS_PER_CELL_UPDATE, aSwitch );  

	  /* update input action */
	  for(input=0; input<aSwitch->numInputs; input++)
	    (aSwitch->inputAction)(INPUTACTION_PER_CELL_UPDATE, aSwitch, 
				   input, NULL);
	}
		

#ifdef _SIM_

      /**********************************************************/
      /******** Poll SimGraph stub (SG)                 *********/
      /**********************************************************/
      for( switchNumber=0; switchNumber<numSwitches && simGraphEnabled
         ; switchNumber++)
    {
      struct Element *anElement;

      aSwitch = switches[switchNumber];

      if( !(now%simGraphSocketPeriod) ) {
        graph_check_socket_ready( NULL, &graphTool, NULL);
      }

      if( now == simGraphPauseTime ) {
        graph_pause_on_socket( NULL, &graphTool, NULL);
      }

      for (anElement = graphTool.listOfRequests->head; anElement;
           anElement = anElement->next) {
        struct GraphRequestPacket *pkt =
          (struct GraphRequestPacket *)anElement->Object;
        struct GraphPeriodReqPktContents *pkt2 =
          &pkt->contents.period;
        if (pkt->reqType == GRAPH_PERIODIC_REPORT) {
          if( !(now%pkt2->period) ) {
          graph_periodicReporting(NULL, &graphTool, pkt);
          }
        }
      }
    }

#endif // _SIM_


      /**********************************************************/
      /*********** See if end condition has been met ************/
      /********  Currently based on mean cell latency ***********/
      /**********************************************************/
      if( now%CHECK_STOP == 0 )
	{
	  if( checkStopCondition() ) 
	    simStopped = STOP_SIMULATION;
	  if( pflag )	
	    fprintf(stderr, "now=%ld, %lu ", 
		    now, simulationLength);
	}

      /**********************************************************/
      /*********** See if memory usage is too high **************/
      /**********************************************************/
      if( now%CHECK_MEMORY_USAGE == 0 )
	{
	  memUsed = memoryUsage();
	  if( memUsed > memMaxUsed )
	    memMaxUsed = memUsed;
	  /*if( memUsed > MAX_ALLOWED_MEMORY_USAGE )
	    {
	      simStopped = STOP_SIMULATION;
	      printf("sim: too much memory allocated: %d bytes\n", memUsed);
	    }*/
	}
		
    }

  /***********************************************************/
  /**************** SIMULATION HAS COMPLETED *****************/
  /***********************************************************/
  printf("\n\n\n");
  printf("===============================================================\n");
  printf("======================== RESULTS ==============================\n");
  printf("===============================================================\n");
  printf("Simulation stopped at time: %lu\n", now);
  printf("Maximum memory used: %d bytes\n", memMaxUsed);
  time(&stopTime);
  printf("# Simulation runtime: %lu secs\n\n", (unsigned long) timeDiff(stopTime, startTime));

  for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
    {
      printf("====================================================\n");
      printf("====================== SWITCH %4d =================\n", 
	     switchNumber);
      printf("====================================================\n");
      aSwitch = switches[switchNumber];
      /* Information about traffic */
      printf("  Traffic Information\n");
      printf("  -------------------\n");
      printf("  I/P  model  Util\n");
      for(input=0; input<aSwitch->numInputs; input++)
	{
	  (aSwitch->inputBuffer[input]->trafficModel)
	    (REPORT_TRAFFIC_STATS, aSwitch, input);
	}
      printf("\n");

      /* Print stats from input action(s). */
      (aSwitch->inputAction)(INPUTACTION_STATS_PRINT, aSwitch, 
			     NULL, NULL);

      switchStats(SWITCH_STATS_PRINT_ALL, aSwitch ); 
      (aSwitch->scheduler.schedulingAlgorithm)
	(SCHEDULING_REPORT_STATS, aSwitch);
      latencyStats(LATENCY_STATS_SWITCH_PRINT, aSwitch, NULL);
      (aSwitch->fabric.fabricAction)(FABRIC_STATS_PRINT, aSwitch);

      printf("\n");
      printf("  OUTPUT BURSTINESS\n");
      printf("  -----------------\n");
      printf("   O/P  Avg   SD   \n");
      printf("  -----------------\n");
      for( output=0; output<aSwitch->numOutputs;output++)
	burstStats(BURST_STATS_PRINT, aSwitch, output);
    }

  latencyStats(LATENCY_STATS_CELL_PRINT, NULL, NULL);
  
	for( switchNumber=0; switchNumber<numSwitches; switchNumber++)
		fprintf(stdout, "Switch %4d: \t [%.3g] \n", switchNumber, (double)((double)(switches[switchNumber]->cellsOut/(now-resetStatsTime))/switches[switchNumber]->numInputs) ) ;
	printf("-----------------------------\n");
	// if( use_gnuplot )	automateLoadDelayCurve( aSwitch ) ;
	
#ifdef _SIM_

    /************************************************************/
    /***** Shutdown  SIMGRAPH  (SG)                  ************/
    /************************************************************/

  if( simGraphEnabled )
        graph_end( &graphTool );

#endif // _SIM_



	free(switches); /* Added Sundar */
  exit(0);
}


/**********************************************************************/
/********************* Support Routines *******************************/
/**********************************************************************/
static int checkStopCondition()
{
  /* 
     Currently: 
     if mean delay of cells changes 
     less than STOP_THRESHOLD in a CHECK_STOP period, 
     then return STOP_SIMULATION;
     else return CONTINUE_SIMULATION;
	*/

  static double lastValue=0.0;
  static int numTimesMetStopCondition=0;
  double thisValue;
  double diff;

  thisValue = latencyStats(LATENCY_STATS_RETURN_AVG, NULL, NULL);
  if( thisValue == 0 )
    return(NO);
  diff = fabs((lastValue - thisValue) / thisValue );
  if( diff < STOP_THRESHOLD )
    numTimesMetStopCondition++;
  else
    numTimesMetStopCondition = 0;
  if( numTimesMetStopCondition >= NUM_TIMES_MUST_MEET_STOP_CONDITION )
    {
      printf("Stopped because converged at time: %ld\n", now);
      return(STOP_SIMULATION);
    }
  else
    {
      lastValue = thisValue;
      return(NO);
    }
		
}

/*************************************************************/
static int memoryUsage()
{
  return(getAllocAmount());
}

/**************************************************************/
static void headerInformation(argc, argv)
  int argc;
char **argv;
{
  int i;
  char hostname[MAXSTRING];
  time_t aTime;

  printf("##################################################\n");
  for(i=0;i<argc;i++)
    printf("%s ", argv[i]); 
  time(&aTime);
  printf("This is Sim version 2.35, Released Sun April 8th 2001 \n"); 
  printf("\n# Date %s", ctime(&aTime)); 

  gethostname(hostname, MAXSTRING);
  printf("# Machine %s\n", hostname);
  printf("##################################################\n");
}

/***************************************************************/
void FatalError(aString)
  char *aString;
{
  fprintf(stderr, "sim: Fatal Error at time %ld.\n", now);
  fprintf(stderr, "sim: %s.\n", aString);
  exit(1);
}

/***************************************************************/

time_t timeDiff(time2, time1)
  time_t time2, time1;
{
  return(time2-time1);
}

/***************************************************************/

void automateLoadDelayCurve( Switch *aSwitch )
{	
	FILE	*load_latency ;
	char	*load_latency_file = "/Users/n_skalis/iCode/Xcode/iSIM/src/_debug/_dats/load_VS_latency.dat" ;

	if((load_latency = fopen(load_latency_file, "a")) == NULL)		
		{ fprintf(stderr, "ERROR : Cannot open file 'load_VS_latency.dat' for writing ... \n") ; fprintf(stderr, "NOTE : 'automateLoadDelayCurve()' function is based on the filename - '%s' \n", load_latency_file) ; exit( EXIT_FAILURE ) ; }
	// fprintf(load_latency, "%8.8g \t %8.8g \n", (double)(inputLoad/aSwitch->numInputs), latencyStats(LATENCY_STATS_RETURN_AVG, NULL, NULL)) ;
	if(fclose( load_latency ))	{ fprintf(stderr, "ERROR : Cannot close file 'load_VS_latency.dat' ... \n") ; }
	
	return ;
}